@@ -164,6 +164,16 @@ void KonquerorApplication::setupAboutData()
     m_aboutData.setCopyrightStatement(i18n("(C) 1999-2023, The Konqueror developers"));
     m_aboutData.setHomepage("https://apps.kde.org/konqueror");
 
+    // The window icon, retrieved in KAboutApplicationDialogPrivate::init(),
+    // may vary for this application depending on what is displayed.  Set the
+    // program logo explicitly for the "About Konqueror" dialogue.
+    m_aboutData.setProgramLogo(QIcon::fromTheme("konqueror"));
+
+    // However, the icon for the "About <application>" action, generated
+    // in KStandardAction::_k_createInternal(), uses the window icon directly.
+    // This may again change later, but set it here for the action.
+    QGuiApplication::setWindowIcon(QIcon::fromTheme("konqueror"));
+
     m_aboutData.addAuthor(i18n("Stefano Crocco"), i18n("Current maintainer"), "stefano.crocco@alice.it");
     m_aboutData.addAuthor(i18n("David Faure"), i18n("Developer (framework, parts, JavaScript, I/O library) and former maintainer"), "faure@kde.org");
     m_aboutData.addAuthor(i18n("Simon Hausmann"), i18n("Developer (framework, parts)"), "hausmann@kde.org");
@@ -344,10 +354,11 @@ int KonquerorApplication::performStart(const QString& workingDirectory, bool fir
             KMessageBox::error(nullptr, i18nc("The session asked by the user doesn't exist or can't be opened",
                                               "Session %1 couldn't be opened", sessionName));
         }
-        //If there was an error opening the session and this is the first instance, don't return immediately
-        //as this would lead to the application being run but with no windows open. Instead, open an empty
+        //Return except when this is the first instance and there was an error loading the session,
+        //since in this case, returning would lead to the application being run with no windows. Instead
+        //go on with the function so an empty window will be created
         //window
-        if (result != 0 && firstInstance) {
+        if (result == 0 || !firstInstance) {
             return result;
         }
     }
