@@ -15,7 +15,7 @@
 
 #include "libkonq_utils.h"
 
-#include "interfaces/downloaderextension.h"
+#include "interfaces/downloadjob.h"
 
 #include <KIO/OpenUrlJob>
 #include <KIO/JobUiDelegate>
@@ -47,26 +47,7 @@
 #include <QLoggingCategory>
 
 using namespace KonqInterfaces;
-
-bool UrlLoader::embedWithoutAskingToSave(const QString &mimeType)
-{
-    static QStringList s_mimeTypes;
-    if (s_mimeTypes.isEmpty()) {
-        QStringList names{QStringLiteral("kfmclient_html"), QStringLiteral("kfmclient_dir"), QStringLiteral("kfmclient_war")};
-        for (const QString &name : names) {
-            KService::Ptr s = KService::serviceByStorageId(name);
-            if (s) {
-                s_mimeTypes.append(s->mimeTypes());
-            } else {
-                qCDebug(KONQUEROR_LOG()) << "Couldn't find service" << name;
-            }
-        }
-        //The user may want to save xml files rather than embedding them
-        //TODO: is there a better way to do this?
-        s_mimeTypes.removeOne(QStringLiteral("application/xml"));
-    }
-    return s_mimeTypes.contains(mimeType);
-}
+using namespace Konq;
 
 UrlLoader::UrlLoader(KonqMainWindow *mainWindow, KonqView *view, const QUrl &url, const QString &mimeType, const KonqOpenURLRequest &req, bool trustedSource):
     QObject(mainWindow),
@@ -77,93 +58,61 @@ UrlLoader::UrlLoader(KonqMainWindow *mainWindow, KonqView *view, const QUrl &url
     m_view(view),
     m_trustedSource(trustedSource),
     m_protocolAllowsReading(KProtocolManager::supportsReading(m_url) || !KProtocolInfo::isKnownProtocol(m_url)), // If the protocol is unknown, assume it allows reading
-    m_letRequestingPartDownloadUrl(req.letPartPerformDownload)
+    m_useDownloadJob(req.browserArgs.downloadJob()),
+    m_downloadJob(req.browserArgs.downloadJob()),
+    m_allowedActions{m_request.urlActions()}
 {
     m_request.suggestedFileName = m_request.browserArgs.suggestedDownloadName();
 
-    //TODO KF6: when dropping compatibility with KF5, stop using metadata and add appropriate fields to BrowserArguments
-    QString requestedEmbeddingPart = m_request.args.metaData().value(QStringLiteral("embed-with"));
+    QString requestedEmbeddingPart = m_request.browserArgs.embedWith();
     if (!requestedEmbeddingPart.isEmpty()) {
         m_request.serviceName = requestedEmbeddingPart;
     }
 
-    QString openWith = m_request.args.metaData().value(QStringLiteral("open-with"));
-    m_request.forceOpen = !openWith.isEmpty();
-    initAllowedActions(m_request);
-
-    //If we're asked to use a specific application to open the URL, assume the user
-    //has already decided to open the URL
-    if (!openWith.isEmpty()) {
-        m_service = KService::serviceByStorageId(openWith);
-    }
-
-    getDownloaderJobFromPart();
-    if (m_letRequestingPartDownloadUrl) {
+    if (m_useDownloadJob) {
         m_originalUrl = m_url;
+        m_downloadJob->setParent(this);
     }
     determineStartingMimetype();
-    m_dontPassToWebEnginePart = m_request.args.metaData().contains("DontSendToDefaultHTMLPart");
+    m_ignoreDefaultHtmlPart = m_request.browserArgs.ignoreDefaultHtmlPart();
 }
 
 UrlLoader::~UrlLoader()
 {
 }
 
-void UrlLoader::initAllowedActions(const KonqOpenURLRequest &req)
+void UrlLoader::updateAllowedActions()
 {
-    QString actionName = req.args.metaData().value("action", {});
-
-    //We have two conflicting requests: do nothing (of course, this should never happen)
-    if (req.forceAutoEmbed && req.forceOpen) {
-        m_allowedActions.clear();
-        m_action = OpenUrlAction::DoNothing;
-        return;
+    if (m_allowedActions.isAllowed(UrlAction::Execute) && !isUrlExecutable()) {
+        m_allowedActions.allow(UrlAction::Execute, false);
     }
 
-    if (req.forceAutoEmbed) {
-        m_allowedActions = {OpenUrlAction::Embed};
-    } else if (req.forceOpen) {
-        m_allowedActions = {OpenUrlAction::Open};
-    } else if (!actionName.isEmpty()) {
-        static QHash<QString, UrlLoader::OpenUrlAction> actions {
-            {QStringLiteral("nothing"), UrlLoader::OpenUrlAction::DoNothing},
-            {QStringLiteral("save"), UrlLoader::OpenUrlAction::Save},
-            {QStringLiteral("embed"), UrlLoader::OpenUrlAction::Embed},
-            {QStringLiteral("open"), UrlLoader::OpenUrlAction::Open},
-            {QStringLiteral("execute"), UrlLoader::OpenUrlAction::Execute}
-        };
-        OpenUrlAction action = actions.value(actionName, UrlLoader::OpenUrlAction::UnknwonAction);
-        if (action == OpenUrlAction::DoNothing) {
-            m_allowedActions.clear();
-        } else if (action != OpenUrlAction::UnknwonAction) {
-            m_allowedActions = {action};
+    if (m_allowedActions.isAllowed(UrlAction::Embed)) {
+        if (!m_part.isValid()) {
+            m_allowedActions.allow(UrlAction::Embed, false);
+        } else if (isViewLocked()) {
+        //If the view is locked, assume, we need to embed the URL. Ideally, this should
+        //be done
+            m_request.forceEmbed();
         }
     }
-}
 
-bool UrlLoader::isForced(OpenUrlAction action) const
-{
-    switch (action) {
-        case OpenUrlAction::UnknwonAction: return false;
-        case OpenUrlAction::DoNothing: return m_allowedActions.isEmpty();
-        default: return m_allowedActions.length() == 1 && m_allowedActions.value(0) == action;
+    if (serviceIsKonqueror(m_service)) {
+        m_allowedActions.allow(UrlAction::Open, false);
     }
+
 }
 
 bool UrlLoader::can(OpenUrlAction action) const
 {
-    if (action == OpenUrlAction::DoNothing || action == OpenUrlAction::UnknwonAction) {
-        return true;
-    }
-    return m_allowedActions.contains(action);
+    return m_allowedActions.isAllowed(action);
 }
 
-
 void UrlLoader::determineStartingMimetype()
 {
     QMimeDatabase db;
     if (!m_mimeType.isEmpty()) {
-        if (m_letRequestingPartDownloadUrl || !db.mimeTypeForName(m_mimeType).isDefault()) {
+        if (m_useDownloadJob || !db.mimeTypeForName(m_mimeType).isDefault()) {
             return;
         }
     }
@@ -171,7 +120,7 @@ void UrlLoader::determineStartingMimetype()
     m_mimeType = m_request.args.mimeType();
     //In theory, this should never happen, as parts requesting to download the URL
     //by themselves should already have set the mimetype. However, let's check, just in case
-    if (m_letRequestingPartDownloadUrl && m_mimeType.isEmpty()) {
+    if (m_useDownloadJob && m_mimeType.isEmpty()) {
         m_mimeType = QStringLiteral("application/octet-stream");
     } else if (db.mimeTypeForName(m_mimeType).isDefault()) {
         m_mimeType.clear();
@@ -203,6 +152,22 @@ void UrlLoader::setNewTab(bool newTab)
     m_request.browserArgs.setNewTab(newTab);
 }
 
+
+void UrlLoader::findService()
+{
+    if (!can(UrlAction::Open)) {
+        m_service = nullptr;
+        return;
+    }
+    QString openWith = m_request.browserArgs.openWith();
+    if (!openWith.isEmpty()) {
+        m_service = KService::serviceByStorageId(openWith);
+    }
+    if (!m_service) {
+        m_service = KApplicationTrader::preferredService(m_mimeType);
+    }
+}
+
 void UrlLoader::start()
 {
     if (m_url.isLocalFile()) {
@@ -214,14 +179,8 @@ void UrlLoader::start()
     if (hasError()) {
         m_mimeType = QStringLiteral("text/html");
     }
-    if (isMimeTypeKnown(m_mimeType)) {
-        KService::Ptr preferredService = KApplicationTrader::preferredService(m_mimeType);
-        if (serviceIsKonqueror(preferredService)) {
-            m_request.forceAutoEmbed = true;
-        }
-    }
 
-    m_isAsync = m_protocolAllowsReading && (!isMimeTypeKnown(m_mimeType) || m_letRequestingPartDownloadUrl);
+    m_isAsync = m_protocolAllowsReading && (!isMimeTypeKnown(m_mimeType) || m_useDownloadJob);
 }
 
 bool UrlLoader::isViewLocked() const
@@ -229,23 +188,21 @@ bool UrlLoader::isViewLocked() const
     return m_view && m_view->isLockedLocation();
 }
 
-bool UrlLoader::setAction(OpenUrlAction action)
-{
-    if (!m_allowedActions.contains(action)) {
-        return false;
-    }
-    m_action = action;
-    return true;
-}
-
 void UrlLoader::decideAction()
 {
     if (hasError()) {
-        setAction(OpenUrlAction::Embed);
+        m_request.forceEmbed();
         return;
     }
-    m_action = decideExecute();
-    switch (m_action) {
+
+    if (isMimeTypeKnown(m_mimeType)) {
+        m_part = findEmbeddingPart();
+        findService();
+        updateAllowedActions();
+    }
+
+    decideExecute();
+    switch (m_request.chosenAction) {
         case OpenUrlAction::Execute:
             m_ready = true;
             break;
@@ -253,18 +210,7 @@ void UrlLoader::decideAction()
             m_ready = true;
             return;
         default:
-            if (!isMimeTypeKnown(m_mimeType) && !m_protocolAllowsReading) {
-                //If the protocol doesn't allow reading and we don't have a mimetype associated with it,
-                //use the Open action, as we most likely won't be able to find out the mimetype. This is
-                //what happens, for example, for mailto URLs
-                setAction(OpenUrlAction::Open);
-                return;
-            } else if (isViewLocked() || shouldEmbedThis()) {
-                if (decideEmbedOrSave()) {
-                    return;
-                }
-            }
-            decideOpenOrSave();
+            decideEmbedOpenOrSave();
     }
 }
 
@@ -286,7 +232,7 @@ void UrlLoader::goOn()
         launchMimeTypeFinderJob();
     } else {
         decideAction();
-        m_ready = !m_letRequestingPartDownloadUrl;
+        m_ready = !m_useDownloadJob;
         performAction();
     }
 }
@@ -338,7 +284,7 @@ KPluginMetaData UrlLoader::findEmbeddingPart(bool forceServiceName) const
     /* Corner case: webenginepart can't determine mimetype (gives application/octet-stream) but
      * OpenUrlJob determines a mimetype supported by WebEnginePart (for example application/xml):
      * if the preferred part is webenginepart, we'd get an endless loop because webenginepart will
-     * call again this. To avoid this, if the preferred service is webenginepart and m_dontPassToWebEnginePart
+     * call again this. To avoid this, if the preferred service is webenginepart and m_ignoreDefaultHtmlPart
      * is true, use the second preferred service (if any); otherwise return false. This will offer the user
      * the option to open or save, instead.
      *
@@ -347,7 +293,7 @@ KPluginMetaData UrlLoader::findEmbeddingPart(bool forceServiceName) const
      * the URL and will ask Konqueror to download it. However, if the preferred part for the URL mimetype is
      * WebEnginePart itself, this would lead to an endless loop. This check avoids it
      */
-    if (m_dontPassToWebEnginePart && m_part.pluginId() == webEngineName) {
+    if (m_ignoreDefaultHtmlPart && m_part.pluginId() == webEngineName) {
         QVector<KPluginMetaData> parts = KParts::PartLoader::partsForMimeType(m_mimeType);
         auto findPart = [&webEngineName](const KPluginMetaData &md){return md.pluginId() != webEngineName;};
         QVector<KPluginMetaData>::const_iterator partToUse = std::find_if(parts.constBegin(), parts.constEnd(), findPart);
@@ -360,72 +306,68 @@ KPluginMetaData UrlLoader::findEmbeddingPart(bool forceServiceName) const
     return part;
 }
 
-bool UrlLoader::decideEmbedOrSave()
+static UrlLoader::OpenUrlAction actionFromDialogResult(DownloadActionQuestion::Action res)
 {
-    m_part = findEmbeddingPart();
-
-    //If we can't find a part, return false, so that the caller can use decideOpenOrSave to allow the
-    //user the possibility of opening the file, since embedding wasn't possibile
-    if (!m_part.isValid()) {
-        return false;
+    switch (res) {
+        case DownloadActionQuestion::Action::Cancel:
+            return UrlLoader::OpenUrlAction::DoNothing;
+        case DownloadActionQuestion::Action::Save:
+            return UrlLoader::OpenUrlAction::Save;
+        case DownloadActionQuestion::Action::Embed:
+            return UrlLoader::OpenUrlAction::Embed;
+        case DownloadActionQuestion::Action::Open:
+            return UrlLoader::OpenUrlAction::Open;
     }
-
-    //Ask whether to save or embed, except in the following cases:
-    //- it's a web page: always embed
-    //- it's a local file: always embed
-    if (embedWithoutAskingToSave(m_mimeType) || m_url.isLocalFile() || isForced(OpenUrlAction::Embed) || !can(OpenUrlAction::Save)) {
-        setAction(OpenUrlAction::Embed);
-    } else {
-        setAction(askSaveOrOpen(OpenEmbedMode::Embed).first);
-    }
-
-    if (m_action == OpenUrlAction::Embed) {
-        m_request.serviceName = m_part.pluginId();
-    }
-
-    m_ready = m_part.isValid() || m_action != OpenUrlAction::Embed;
-    return true;
+    //We can't really reach this point, but it's needed to avoid a compiler warning
+    return UrlLoader::OpenUrlAction::UnknownAction;
 }
 
-void UrlLoader::decideOpenOrSave()
+void UrlLoader::decideEmbedOpenOrSave()
 {
-    m_ready = true;
-    if (!can(OpenUrlAction::Save) && !can(OpenUrlAction::Open)) {
-        //decideOpenOrSave is the last option. If we can't neither open nor save, do nothing
-        setAction(OpenUrlAction::DoNothing);
-        return;
-    }
-    QString protClass = KProtocolInfo::protocolClass(m_url.scheme());
-    bool isLocal = m_url.isLocalFile();
-    bool alwaysOpen = isLocal || protClass == QLatin1String(":local") || KProtocolInfo::isHelperProtocol(m_url);
-    OpenSaveAnswer answerWithService;
-    if (alwaysOpen || isForced(OpenUrlAction::Open) || !can(OpenUrlAction::Save)) {
-        answerWithService = qMakePair(OpenUrlAction::Open, KApplicationTrader::preferredService(m_mimeType));
-    } else if (isForced(OpenUrlAction::Save) || !can(OpenUrlAction::Open)) {
-        answerWithService = qMakePair(OpenUrlAction::Save, nullptr);
-    } else {
-        answerWithService = askSaveOrOpen(OpenEmbedMode::Open);
+    m_request.chosenAction = m_allowedActions.forcedAction();
+    if (m_request.chosenAction == OpenUrlAction::UnknownAction) {
+        askEmbedSaveOrOpen();
     }
 
-    setAction(answerWithService.first);
-
-    //If the URL should be opened in an external application and it should be downloaded by the requesting part,
+    //If the URL should be opened in an external application and it should be downloaded using a download job
     //ensure it's not downloaded in Konqueror's temporary directory but in the global temporary directory,
     //otherwise if the user closes Konqueror before closing the application, there could be issues because the
     //external application wouldn't find the file anymore.
     //Problem: if the application doesn't support the --tempfile switch, the file will remain even after both Konqueror
     //and the external application are closed and will only be removed automatically if the temporary directory is deleted
     //or emptied.
-    if (m_letRequestingPartDownloadUrl && m_action == OpenUrlAction::Open && m_partDownloaderJob) {
-        QString fileName = QFileInfo(m_partDownloaderJob->downloadPath()).fileName();
-        fileName = Konq::generateUniqueFileName(fileName, QDir::temp().path());
-        m_partDownloaderJob->setDownloadPath(QDir::temp().filePath(fileName));
-    }
-    if (!m_service) {
-        m_service = answerWithService.second;
+    DownloadJob *job = m_request.browserArgs.downloadJob();
+    if (m_useDownloadJob && m_request.chosenAction == OpenUrlAction::Open && job) {
+        QString fileName = QFileInfo(job->downloadPath()).fileName();
+        fileName = generateUniqueFileName(fileName, QDir::temp().path());
+        job->setDownloadPath(QDir::temp().filePath(fileName));
     }
+
+    m_ready = m_request.chosenAction != OpenUrlAction::Embed || m_part.isValid();
 }
 
+void UrlLoader::askEmbedSaveOrOpen()
+{
+    using Result = DownloadActionQuestion::Action;
+    DownloadActionQuestion dlg(m_mainWindow, m_url, m_mimeType, m_ignoreDefaultHtmlPart);
+    dlg.setSuggestedFileName(m_request.suggestedFileName);
+    Result saveMode = can(OpenUrlAction::Save) ? Result::Save : Result::Cancel;
+    Result openMode = can(OpenUrlAction::Open) ? Result::Open : Result::Cancel;
+    Result embedMode = can(OpenUrlAction::Embed) ? Result::Embed : Result::Cancel;
+
+    DownloadActionQuestion::EmbedFlags flag = m_request.browserArgs.forceShowActionDialog() ? DownloadActionQuestion::ForceDialog : DownloadActionQuestion::InlineDisposition;
+    DownloadActionQuestion::Action ans = dlg.ask(saveMode|embedMode|openMode, flag);
+
+    m_request.chosenAction = actionFromDialogResult(ans);
+    if (m_request.chosenAction == OpenUrlAction::Embed && (m_request.browserArgs.embedWith().isEmpty() || dlg.dialogShown())) {
+        //Override the part in browserArgs only if the user was shown a dialog (meaning he explicitly chose a part)
+        m_part = dlg.selectedPart();
+        m_request.serviceName = m_part.pluginId();
+    } else if (m_request.chosenAction == OpenUrlAction::Open && (!m_request.browserArgs.openWith().isEmpty() || dlg.dialogShown())) {
+        //Override the service in browserArgs only if the user was shown a dialog (meaning he explicitly chose a service)
+        m_service = dlg.selectedService();
+    }
+}
 
 bool UrlLoader::isUrlExecutable() const
 {
@@ -453,23 +395,37 @@ bool UrlLoader::isUrlExecutable() const
 }
 
 
-UrlLoader::OpenUrlAction UrlLoader::decideExecute() const {
-    //We don't want to execute files which aren't local, files which aren't executable (obviously)
-    //and we don't want to execute when we are reloading (the file is visible in the current part,
-    //so we know the user wanted to display it. We also don't want to execute when we aren't allowed to do so
-    if (!can(OpenUrlAction::Execute) || !isUrlExecutable() || m_request.args.reload()) {
-        return OpenUrlAction::UnknwonAction;
+void UrlLoader::decideExecute()
+{
+    if (!can(UrlAction::Execute)) {
+        return;
+    }
+
+    //We don't want to execute when we are reloading the file is visible in the current part
+    //(the file is visible in the current part, so we know the user wanted to display it) unless
+    //for, some reasons, we're forced to do so.
+    if (m_request.args.reload() && !m_allowedActions.isForced(UrlAction::Execute)) {
+        return;
+    }
+
+    UrlAction alternativeAction;
+    if (can(UrlAction::Embed)) {
+        alternativeAction = UrlAction::Embed;
+    } else if (can(UrlAction::Open)) {
+        alternativeAction = UrlAction::Open;
+    } else {
+        alternativeAction = UrlAction::DoNothing;
     }
-    bool canDisplay = !KParts::PartLoader::partsForMimeType(m_mimeType).isEmpty();
 
     KMessageBox::ButtonCode code;
     KGuiItem executeGuiItem(i18nc("Execute an executable file", "Execute"),
                             QIcon::fromTheme(QStringLiteral("system-run")));
-    KGuiItem displayGuiItem(i18nc("Display an executable file", "Display"),
-                            QIcon::fromTheme(QStringLiteral("document-preview")));
+    QString displayIconName = alternativeAction == UrlAction::Embed ? QStringLiteral("document-preview") : QStringLiteral("document-open");
+    KGuiItem displayGuiItem(i18nc("Display an executable file inside Konqueror", "Display"),
+                            QIcon::fromTheme(displayIconName));
     QString dontShowAgainId(QLatin1String("AskExecuting")+m_mimeType);
 
-    if (canDisplay) {
+    if (alternativeAction != UrlAction::DoNothing) {
         code = KMessageBox::questionTwoActionsCancel(m_mainWindow,
                                                      xi18nc("@info The user has to decide whether to execute an executable file or display it",
                                                             "<filename>%1</filename> can be executed. Do you want to execute it or to display it?", m_url.path()),
@@ -483,31 +439,32 @@ UrlLoader::OpenUrlAction UrlLoader::decideExecute() const {
                                                dontShowAgainId, KMessageBox::Dangerous);}
     switch (code) {
         case KMessageBox::PrimaryAction:
-            return OpenUrlAction::Execute;
+            m_request.chosenAction = OpenUrlAction::Execute;
+            break;
         case KMessageBox::Cancel:
-            return OpenUrlAction::DoNothing;
+            m_request.chosenAction = OpenUrlAction::DoNothing;
+            break;
         case KMessageBox::SecondaryAction:
             //The "No" button actually corresponds to the "Cancel" action if the file can't be displayed
-            return canDisplay ? OpenUrlAction::UnknwonAction : OpenUrlAction::DoNothing;
-        default: //This is here only to avoid a compiler warning
-            return OpenUrlAction::UnknwonAction;
+            m_request.chosenAction = alternativeAction;
+        default: //This point can't be reached. This is here only to avoid a compiler warning
+            return;
     }
 }
 
 void UrlLoader::performAction()
 {
-    //If we still aren't ready, it means that the part wants to download the URL
-    //by itself. Do this only when opening or embedding, however, since when
+    //If we still aren't ready, it means that we need to use a download job.
+    //Do this only when opening or embedding, however, since when
     //saving we first need to ask the user where to save. save() will launch the
     //job itself.
-    //When the part has finished downloading the URL, the slot connected with the
-    //job's result() signal will call again this function, after setting m_ready
-    //to true
-    if (!m_ready && (m_action == OpenUrlAction::Embed || m_action == OpenUrlAction::Open)) {
+    //When the job has finished downloading the URL, the slot connected with its
+    //result() signal will set m_redy to true, then call this function again
+    if (!m_ready && (m_request.chosenAction == OpenUrlAction::Embed || m_request.chosenAction == OpenUrlAction::Open)) {
         downloadForEmbeddingOrOpening();
         return;
     }
-    switch (m_action) {
+    switch (m_request.chosenAction) {
         case OpenUrlAction::Embed:
             embed();
             break;
@@ -521,53 +478,32 @@ void UrlLoader::performAction()
             save();
             break;
         case OpenUrlAction::DoNothing:
-        case OpenUrlAction::UnknwonAction: //This should never happen
+        case OpenUrlAction::UnknownAction: //This should never happen
             done();
             break;
     }
 }
 
-void UrlLoader::getDownloaderJobFromPart()
-{
-    if (!m_letRequestingPartDownloadUrl) {
-        return;
-    }
-    DownloaderExtension *iface = downloaderInterface();
-    if (iface) {
-        m_partDownloaderJob = iface->downloadJob(m_url, m_request.downloadId, this);
-    } else {
-        qCDebug(KONQUEROR_LOG) << "Wanting to let part download" << m_url << "but part doesn't implement the DownloaderInterface";
-    }
-
-    //If we can't get a job for whatever reason (it shouldn't happen, but let's be sure)
-    //try to open the URL in the usual way. Maybe it'll work (most likely, it will if
-    //there's no need to have special cookies set to access it)
-    if (!m_partDownloaderJob) {
-        qCDebug(KONQUEROR_LOG) << "Couldn't get DownloadJob for" << m_url << "from part" << m_part;
-        m_letRequestingPartDownloadUrl = false;
-        m_request.tempFile = false; //Opening a remote URL with tempFile will fail
-    }
-}
-
 void UrlLoader::downloadForEmbeddingOrOpening()
 {
     //This shouldn't happen
-    if (!m_partDownloaderJob) {
+    if (!m_downloadJob) {
         done();
         return;
     }
-    connect(m_partDownloaderJob, &DownloaderJob::downloadResult, this, &UrlLoader::jobFinished);
-    m_partDownloaderJob->startDownload(m_mainWindow, this, &UrlLoader::downloadForEmbeddingOrOpeningDone);
+    m_downloadJob->setIntent(m_request.chosenAction == UrlAction::Embed ? DownloadJob::Embed : DownloadJob::Open);
+    connect(m_downloadJob, &DownloadJob::downloadResult, this, &UrlLoader::jobFinished);
+    m_downloadJob->startDownload(m_mainWindow, this, &UrlLoader::downloadForEmbeddingOrOpeningDone);
 }
 
-void UrlLoader::downloadForEmbeddingOrOpeningDone(KonqInterfaces::DownloaderJob *job, const QUrl &url)
+void UrlLoader::downloadForEmbeddingOrOpeningDone(KonqInterfaces::DownloadJob *job, const QUrl &url)
 {
     if (job && job->error() == 0) {
         m_url = url;
         m_ready = true;
         m_request.tempFile = true;
     } else if (!job || job->error() == KIO::ERR_USER_CANCELED) {
-        m_action = OpenUrlAction::DoNothing;
+        m_request.chosenAction = OpenUrlAction::DoNothing;
         m_ready = true;
     }
     checkDownloadedMimetype();
@@ -577,22 +513,23 @@ void UrlLoader::downloadForEmbeddingOrOpeningDone(KonqInterfaces::DownloaderJob
 void UrlLoader::checkDownloadedMimetype()
 {
     QMimeDatabase db;
-    QString type = db.mimeTypeForFile(m_url.path()).name();
+    QMimeType type = db.mimeTypeForFile(m_url.path());
+    QString typeName = type.name();
     //Don't override the mimetype with the default one
-    if (type == m_mimeType || type == "application/octet-stream") {
+    if (type.isDefault() || typeName == m_mimeType) {
         return;
     }
-    m_mimeType = type;
+    m_mimeType = typeName;
     //The URL is a local file now, so there are no problems in opening it with WebEnginePart
-    m_dontPassToWebEnginePart = false;
+    m_ignoreDefaultHtmlPart = false;
     if (shouldEmbedThis()) {
         m_part = findEmbeddingPart(false);
         if (m_part.isValid()) {
-            m_action = OpenUrlAction::Embed;
+            m_request.chosenAction = OpenUrlAction::Embed;
             return;
         }
     }
-    m_action = OpenUrlAction::Open;
+    m_request.chosenAction = OpenUrlAction::Open;
     if (m_service && m_service->hasMimeType(m_mimeType)) {
         return;
     }
@@ -603,14 +540,18 @@ void UrlLoader::done(KJob *job)
 {
     //Ensure that m_mimeType and m_request.args.mimeType are equal, since it's not clear what will be used
     m_request.args.setMimeType(m_mimeType);
+
+    //Update the allowed actions
+    m_request.browserArgs.setAllowedUrlActions(m_allowedActions);
+
     if (job) {
         jobFinished(job);
     }
     emit finished(this);
-    //If we reach here and m_partDownloaderJob->finished() is false, it means the job hasn't been started in the first place,
+    //If we reach here and m_partDownloadJob->finished() is false, it means the job hasn't been started in the first place,
     //(because the user canceled the download), so kill it
-    if (m_partDownloaderJob && !m_partDownloaderJob->finished()) {
-        m_partDownloaderJob->kill();
+    if (m_downloadJob && !m_downloadJob->finished()) {
+        m_downloadJob->kill();
     }
     deleteLater();
 }
@@ -633,9 +574,9 @@ void UrlLoader::mimetypeDeterminedByJob()
 {
     if (m_mimeTypeFinderJob->error()) {
         m_jobErrorCode = m_mimeTypeFinderJob->error();
-        m_url = Konq::makeErrorUrl(m_jobErrorCode, m_mimeTypeFinderJob->errorString(), m_url);
+        m_url = makeErrorUrl(m_jobErrorCode, m_mimeTypeFinderJob->errorString(), m_url);
         m_mimeType = QStringLiteral("text/html");
-        m_action = OpenUrlAction::Embed;
+        m_request.chosenAction = OpenUrlAction::Embed;
         performAction();
         return;
     }
@@ -656,21 +597,13 @@ void UrlLoader::mimetypeDeterminedByJob()
 
 bool UrlLoader::shouldUseDefaultHttpMimeype() const
 {
-    if (m_dontPassToWebEnginePart || isMimeTypeKnown(m_mimeType)) {
+    if (m_ignoreDefaultHtmlPart || isMimeTypeKnown(m_mimeType)) {
         return false;
     }
-    const QVector<QString> webengineSchemes = {QStringLiteral("error"), QStringLiteral("konq")};
+    const QVector<QString> webengineSchemes = {QStringLiteral("error"), QStringLiteral("konq"), QStringLiteral("data")};
     return m_url.scheme().startsWith(QStringLiteral("http")) || webengineSchemes.contains(m_url.scheme());
 }
 
-DownloaderExtension* UrlLoader::downloaderInterface() const
-{
-    if (!m_request.requestingPart) {
-        return nullptr;
-    }
-    return DownloaderExtension::downloader(m_request.requestingPart);
-}
-
 void UrlLoader::detectSettingsForRemoteFiles()
 {
     if (m_url.isLocalFile()) {
@@ -678,13 +611,13 @@ void UrlLoader::detectSettingsForRemoteFiles()
     }
 
     if (m_url.scheme() == QLatin1String("error")) {
-        m_letRequestingPartDownloadUrl = false; //error URLs can never be downloaded
+        m_useDownloadJob = false; //error URLs can never be downloaded
         m_mimeType = QLatin1String("text/html");
         m_request.args.setMimeType(QStringLiteral("text/html"));
     }
     else if (shouldUseDefaultHttpMimeype()) {
         // If a part which supports html asked to download the URL, it means it's not html, so don't change its mimetype
-        if (m_letRequestingPartDownloadUrl && m_part.supportsMimeType(QStringLiteral("text/html"))) {
+        if (m_useDownloadJob && m_part.supportsMimeType(QStringLiteral("text/html"))) {
             return;
         }
         m_mimeType = QLatin1String("text/html");
@@ -750,7 +683,7 @@ void UrlLoader::detectSettingsForLocalFiles()
         return;
     }
 
-    m_letRequestingPartDownloadUrl = false; //If the file is local, there's no need to download it
+    m_useDownloadJob = false; //If the file is local, there's no need to download it
 
     m_jobErrorCode = checkAccessToLocalFile(m_url.path());
 
@@ -777,10 +710,10 @@ void UrlLoader::detectSettingsForLocalFiles()
 
 bool UrlLoader::shouldEmbedThis() const
 {
-    if (!m_allowedActions.contains(OpenUrlAction::Embed)) {
+    if (!can(OpenUrlAction::Embed)) {
         return false;
     } else {
-        return  isForced(OpenUrlAction::Embed) || KonqFMSettings::settings()->shouldEmbed(m_mimeType);
+        return  m_allowedActions.isForced(OpenUrlAction::Embed) || KonqFMSettings::settings()->shouldEmbed(m_mimeType);
     }
 }
 
@@ -788,7 +721,7 @@ void UrlLoader::embed()
 {
     if (m_jobErrorCode) {
         QUrl url = m_url;
-        m_url = Konq::makeErrorUrl(m_jobErrorCode, m_url.scheme(), m_url);
+        m_url = makeErrorUrl(m_jobErrorCode, m_url.scheme(), m_url);
         m_mimeType = QStringLiteral("text/html");
         m_part = findPartById(QStringLiteral("webenginepart"));
     }
@@ -796,7 +729,13 @@ void UrlLoader::embed()
     if (embedded) {
         done();
     } else {
-        decideOpenOrSave();
+        m_allowedActions.allow(UrlAction::Embed, false);
+        m_request.serviceName = {};
+        //This is an unexpected situation, so enable open and save action even if they hadn't been originally requested
+        m_allowedActions.allow(Konq::UrlAction::Open, true);
+        m_allowedActions.allow(Konq::UrlAction::Save, true);
+        m_part = {};
+        decideEmbedOpenOrSave();
         performAction();
     }
 }
@@ -811,7 +750,7 @@ void UrlLoader::save()
     dlg.selectFile(suggestedName);
     dlg.setDirectory(QStandardPaths::writableLocation(QStandardPaths::DownloadLocation));
     //NOTE: we can't use QDialog::open() because the dialog needs to be synchronous so that we can call
-    //DownloaderJob::setDownloadPath()
+    //DownloadJob::setDownloadPath()
     if (dlg.exec() == QDialog::Rejected) {
         done();
         return;
@@ -821,15 +760,16 @@ void UrlLoader::save()
 
 void UrlLoader::performSave(const QUrl& orig, const QUrl& dest)
 {
-    KJob *job = nullptr;
-    if (m_letRequestingPartDownloadUrl) {
-        getDownloaderJobFromPart();
-        if (m_partDownloaderJob) {
-            m_partDownloaderJob->startDownload(dest.path(), m_mainWindow, this, [this](DownloaderJob *j){done(j);});
+    if (m_useDownloadJob) {
+        KonqInterfaces::DownloadJob *dj = m_request.browserArgs.downloadJob();
+        if (dj) {
+            dj->setIntent(DownloadJob::Save);
+            dj->startDownload(dest.path(), m_mainWindow, this, [this](DownloadJob *j){done(j);});
             return;
         }
     }
-    job = KIO::file_copy(orig, dest, -1, KIO::Overwrite);
+
+    KJob *job = KIO::file_copy(orig, dest, -1, KIO::Overwrite);
     if (!job) {
         done();
         return;
@@ -886,29 +826,6 @@ bool UrlLoader::isTextExecutable(const QString &mimeType)
     return ( mimeType == QLatin1String("application/x-desktop") || mimeType == QLatin1String("application/x-shellscript"));
 }
 
-UrlLoader::OpenSaveAnswer UrlLoader::askSaveOrOpen(OpenEmbedMode mode) const
-{
-    BrowserOpenOrSaveQuestion dlg(m_mainWindow, m_url, m_mimeType);
-    dlg.setSuggestedFileName(m_request.suggestedFileName);
-    dlg.setFeatures(BrowserOpenOrSaveQuestion::ServiceSelection);
-    BrowserOpenOrSaveQuestion::Result ans = mode == OpenEmbedMode::Open ? dlg.askOpenOrSave() : dlg.askEmbedOrSave();
-    OpenUrlAction action;
-    switch (ans) {
-        case BrowserOpenOrSaveQuestion::Save:
-            action = OpenUrlAction::Save;
-            break;
-        case BrowserOpenOrSaveQuestion::Open:
-            action = OpenUrlAction::Open;
-            break;
-        case BrowserOpenOrSaveQuestion::Embed:
-            action = OpenUrlAction::Embed;
-            break;
-        default:
-            action = OpenUrlAction::DoNothing;
-    }
-    return qMakePair(action, dlg.selectedService());
-}
-
 QString UrlLoader::partForLocalFile(const QString& path)
 {
     QMimeDatabase db;
@@ -936,31 +853,3 @@ void UrlLoader::jobFinished(KJob* job)
 {
     m_jobErrorCode = job->error();
 }
-
-QDebug operator<<(QDebug dbg, UrlLoader::OpenUrlAction action)
-{
-    QDebugStateSaver saver(dbg);
-    dbg.resetFormat();
-    switch (action) {
-        case UrlLoader::OpenUrlAction::UnknwonAction:
-            dbg << "UnknownAction";
-            break;
-        case UrlLoader::OpenUrlAction::DoNothing:
-            dbg << "DoNothing";
-            break;
-        case UrlLoader::OpenUrlAction::Save:
-            dbg << "Save";
-            break;
-        case UrlLoader::OpenUrlAction::Embed:
-            dbg << "Embed";
-            break;
-        case UrlLoader::OpenUrlAction::Open:
-            dbg << "Open";
-            break;
-        case UrlLoader::OpenUrlAction::Execute:
-            dbg << "Execute";
-            break;
-    }
-    return dbg;
-}
-
